<?php
/*
 * ----------------------------------------------------------------------------
 * Va-Functions.inc.php                                              05/12/2004
 * This script contains a set of functions that will make it easier to get
 * information from the Virtual Advanced configuration and data files. It also
 * contains basic functions to aid you when creating new scripts.
 *
 * This script should always be included first in any script that will use any
 * VADV-PHP functions.
 * ----------------------------------------------------------------------------
 * Copyright (C) 2002-2004 Steve Winn. All Rights Reserved.
 *
 * The source code contains a fair amount of comments, which may aid anyone
 * attempting to follow the program or to modify it.
 *
 * For licensing and copyright information, view LICENSE.TXT.
 * ----------------------------------------------------------------------------
 */

$VERSION    = '040512';                             // Current VADV-PHP version.
$DEBUG_MAIL = FALSE;                                // Turn on/off debug mailer.
                                                    // This should only be used while developing
                                                    // and turned off on a production site.

set_error_handler('CustomErrorHandler');            // Set the default error handler.
error_reporting(E_ERROR |                           // Set errors to be reported,
                E_WARNING |                         // which is used by the error handler.
                E_NOTICE |
                E_USER_ERROR |
                E_USER_WARNING |
                E_USER_NOTICE);

require_once('include_dir.php');                    // Get include directory.
require_once($includeDir . 'common.inc.php');       // Include common settings.
require_once($includeDir . 'va-syspaths.inc.php');  // Include the BBS's paths.
require_once($includeDir . 'va-loginauth.inc.php'); // Login authentication.

$VA_USERS = UpdateWhosOnline();                     // Update the WEBCALLER.LST file.


/**
* Va-Functions.inc.php :: ReadCfgFile()
*
* This function will read the entire file specified into one variable.
*
* @param [type] $file
* @return
*/
function ReadCfgFile($file) {
    if (!is_string($file)) {
        trigger_error(VA_Error(1));
    }

    // Check to see if the file exists, if not return nothing.
    if (!is_file($file)) {
        return '';                          // If file is missing, then it is blank.
    }
    if ((is_file($file)) && (!is_readable($file))) {
        return FALSE;                       // If file exists but cannot be read, return FALSE.
    }

    $fileString = file_get_contents($file);

    return $fileString;                     // Return the file string.
}


/**
* Va-Functions.inc.php :: WriteCfgFile()
*
* This function will create a new configuration file based on the file string
* passed.
*
* @param [type] $file
* @param [type] $fileString
* @return
*/
function WriteCfgFile($file, $fileString) {
    if (!is_string($file)) {
        trigger_error(VA_Error(1));
    }
    if (!is_string($fileString)) {
        trigger_error(VA_Error(1));
    }

    $length = strlen($fileString);          // Determine the length of the file string.

    $fp = fopen($file, 'wb');               // Open the file.
    flock($fp, LOCK_EX);                    // Lock the file.
    
    $status = fwrite($fp, $fileString, $length);    // Write the new file.

    flock($fp, LOCK_UN);                    // Unlock the file.
    fclose($fp);                            // Close the file.

    return $status;                         // Return the write status.
}


/**
* Va-Functions.inc.php :: GetString()
*
* This function will take the given file string and extract the string value
* by using its position and length.
*
* @param [type] $fileString
* @param [type] $position
* @param [type] $length
* @return
*/
function GetString($fileString, $position, $length) {
    if (!is_string($fileString)) {
        trigger_error(VA_Error(1));
    }
    if ((!is_int($position)) && (!is_string($position))) {
        trigger_error(VA_Error(1));
    }
    if ((!is_int($length)) && (!is_string($length))) {
        trigger_error(VA_Error(1));
    }

    // Check to see if the variable was passed, if not then return nothing.
    if ($fileString == '') {
        return '';
    }

    // Check to see if the position is larger than the string length.
    if ($position > strlen($fileString)) {
        return '';
    }

    $position--;                            // Reduce the position by one.
    $string = trim(substr($fileString, $position, $length));

    return strval($string);                 // Return the string.
}


/**
* Va-Functions.inc.php :: GetChar()
*
* This function will take the given file string and extract the character value
* by using its position.
*
* @param [type] $fileString
* @param [type] $position
* @return
*/
function GetChar($fileString, $position) {
    if (!is_string($fileString)) {
        trigger_error(VA_Error(1));
    }
    if ((!is_int($position)) && (!is_string($position))) {
        trigger_error(VA_Error(1));
    }

    // Check to see if the variable was passed, if not then return nothing.
    if ($fileString == '') {
        return '';
    }

    // Check to see if the position is larger than the string length.
    if ($position > strlen($fileString)) {
        return '';
    }

    $position--;                            // Reduce the position by one.
    // Get the first ascii value.
    $a = ord(substr($fileString, $position, 1));

    return intval($a);                      // Return the integer.
}


/**
* Va-Functions.inc.php :: GetInteger()
*
* This function will take the given file string and extract the integer value
* by using its position. If you set $unsigned to 1, then you can extract
* values up to 65535 that are positive only.
*
* @param [type] $fileString
* @param [type] $position
* @param integer $unsigned
* @return
*/
function GetInteger($fileString, $position, $unsigned = 0) {
    if (!is_string($fileString)) {
        trigger_error(VA_Error(1));
    }
    if ((!is_int($position)) && (!is_string($position))) {
        trigger_error(VA_Error(1));
    }

    // Check to see if the variable was passed, if not then return nothing.
    if ($fileString == '') {
        return '';
    }

    // Check to see if the position is larger than the string length.
    if ($position > strlen($fileString)) {
        return '';
    }

    $position--;                            // Reduce the position by one.
    // Get the first ascii value.
    $a = ord(substr($fileString, $position, 1));

    $position++;                            // Move to the next position.
    // Get the second ascii value.
    $b = 256 * ord(substr($fileString, $position, 1));
    
    $integer = intval($a + $b);             // Create the integer.
    
    // Check for negative value.
    if (($unsigned == 0) && ($integer > 32767)) {
        $integer = $integer - 65535 - 1;
    }

    return $integer;                        // Return the integer.
}


/**
* Va-Functions.inc.php :: GetLongInt()
*
* This function will take the given file string and extract the long integer
* value by using its position. If you set $unsigned to 1, then you can extract
* values up to 4294967295 that are positive only.
*
* @param [type] $fileString
* @param [type] $position
* @param integer $unsigned
* @return
*/
function GetLongInt($fileString, $position, $unsigned = 0) {
    if (!is_string($fileString)) {
        trigger_error(VA_Error(1));
    }
    if ((!is_int($position)) && (!is_string($position))) {
        trigger_error(VA_Error(1));
    }

    // Check to see if the variable was passed, if not then return nothing.
    if ($fileString == '') {
        return '';
    }
    
    // Check to see if the position is larger than the string length.
    if ($position > strlen($fileString)) {
        return '';
    }
    
    $position--;                            // Reduce the position by one.
    // Get the first ascii value.
    $a = ord(substr($fileString, $position, 1));
    
    $position++;                            // Move to the next position.
    // Get the second ascii value.
    $b = 256 * ord(substr($fileString, $position, 1));

    $position++;                            // Move to the next position.
    // Get the third ascii value.
    $c = 65536 * ord(substr($fileString, $position, 1));
    
    $position++;                            // Move to the next position.
    // Get the fourth ascii value.
    $d = 16777216 * ord(substr($fileString, $position, 1));
    
    $integer = intval($a + $b + $c + $d);   // Create the integer.
    
    // Check for negative value.
    if (($unsigned == 0) && ($integer > 2147483647)) {
        $integer = $integer - 2147483647 - 1;
    }

    return $integer;                        // Return the long integer.
}


/**
* Va-Functions.inc.php :: GetSingle()
*
* This function will take the given file string and extract the single precison
* value by using its position. If $round is set to any postitive value, then
* the answer will be rounded off to that many decimal places.
*
* This function does not work 100% correctly, especially for larger values. It
* correctly extracts the VADV values, but does not reliably calculate other
* values that may equal or exceed thousands.
*
* Information to do these calculations were taken from:
* http://www.learnbinary.com/binary.html
* http://www.aplawrence.com/Basics/floatingpoint.html
*
* @param [type] $fileString
* @param [type] $position
* @param integer $round
* @return
*/
function GetSingle($fileString, $position, $round = -1) {
    if (!is_string($fileString)) {
        trigger_error(VA_Error(1));
    }
    if ((!is_int($position)) && (!is_string($position))) {
        trigger_error(VA_Error(1));
    }

    // Save the current precision setting...
    $precision = ini_get('precision');
    ini_set('precision', 40);

    // Extract the four bytes, convert to binary and pad them with zeros if needed.
    $a = str_pad(decbin(GetChar($fileString, $position)), 8, 0, STR_PAD_LEFT);
    $b = str_pad(decbin(GetChar($fileString, $position + 1)), 8, 0, STR_PAD_LEFT);
    $c = str_pad(decbin(GetChar($fileString, $position + 2)), 8, 0, STR_PAD_LEFT);
    $d = str_pad(decbin(GetChar($fileString, $position + 3)), 8, 0, STR_PAD_LEFT);

    // Put together the 32 bit binary number.
    $long = $d . $c . $b . $a;

    // Break into separate parts.
    $sign = substr($long, 0, 1);
    $exp  = bindec(substr($long, 1, 8)) - 127;
    $mant = 1 . '.' . substr($long, 9);

    // Calculate the binary answer.
    if ($exp >= 0) {                        // If positive, do the easy way.
        $float = $mant * decbin(pow(2, $exp));
    } else {
        $float = $mant * (10 / (str_pad(1, 2 - $exp, 0, STR_PAD_RIGHT)));
    }

    // If the negative exp is too long, PHP does it in abbreviated notation,
    // so I'm converting it to a string in order to do the rest of the
    // calculations.
    $float = strtoupper($float);
    $pos   = strpos($float, 'E');
    if ($pos !== FALSE) {
        $pol = substr($float, $pos, 1);
        $pow = intval(substr($float, $pos + 1));
        if ($pow < 0) {
            $pow = substr($pow, 1);
        }
        $float = substr($float, 0, $pos);
        $float = str_replace('.', '', $float);
        if ($pol == '-') {
            $float = '0.' . str_pad($float, $pow - 1 + strlen($float), '0', STR_PAD_LEFT);
        } else {
            if ($pow > strlen($float)) {
                $float = str_pad($float, $pow + 1, '0', STR_PAD_RIGHT);
            } else {
                $float = substr($float, 0, $pow) . '.' . substr($float, $pow + 1);
            }
        }
    }
    
    // Split the number into an integer and decimal.
    $array   = explode('.', $float);
    $integer = bindec($array[0]);
    if (!isset($array[1])) {
        $array[1] = '';
    }
    
    // Calculate the fractions of the decimal part (which is in binary).
    // This is the longer, harder way...
    $decimal  = 0;
    $digit    = 1;
    $fraction = 2;
    for ($x = 0; $digit != ''; $x++) {
        $digit = substr($array[1], $x, 1);
        if ($digit != '') {
            $decimal = $decimal + ((1 / $fraction) * $digit);
        }
        $fraction = $fraction * 2;
    }

    // This is the shorter, more compact way...
    // $decimal = bindec($array[1]) / pow(2, strlen($array[1]));

    // Put the two together to get your answer.
    $float = $integer + $decimal;

    // Round the answer if specified.
    if ($round >= 0) {
        $float = round($float, $round);
    } else {
        $float = round($float, 10);
    }

    // Add on the sign if negative.
    if ($sign == 1) {
        $float = '-' . $float;
    }

    // Set the precision back.
    ini_set('precision', $precision);

    return $float;                          // Return the floating point number.
}


/**
* Va-Functions.inc.php :: PutString()
*
* This function will put the passed string into the file string at the given
* position.
*
* @param [type] $fileString
* @param [type] $string
* @param [type] $position
* @param [type] $length
* @return
*/
function PutString($fileString, $string, $position, $length) {
    if (!is_string($fileString)) {
        trigger_error(VA_Error(1));
    }
    if ((!is_int($string)) && (!is_string($string))) {
        trigger_error(VA_Error(1));
    }
    if ((!is_int($position)) && (!is_string($position))) {
        trigger_error(VA_Error(1));
    }
    if ((!is_int($length)) && (!is_string($length))) {
        trigger_error(VA_Error(1));
    }
    
    if (strlen($fileString) < $position) {
       $fileString = $fileString . str_replace('*', chr(0), str_pad('', $position, '*'));
    }

    $position--;

    if (strlen($string) < $length) {
        $string = str_pad($string, $length, chr(32));
    } else {
        $string = substr($string, 0, $length);
    }
    $fileString = substr_replace($fileString, $string, $position, $length);
    
    return $fileString;                     // Return the new string.
}


/**
* Va-Functions.inc.php :: PutChar()
*
* This function will put the passed integer into the file string at the given
* position. This is used for writing characters, or one byte integers.
*
* @param [type] $fileString
* @param [type] $int
* @param [type] $position
* @return
*/
function PutChar($fileString, $int, $position) {
    if (!is_string($fileString)) {
        trigger_error(VA_Error(1));
    }
    if ((!is_int($int)) && (!is_string($position))) {
        trigger_error(VA_Error(1));
    }
    if ((!is_int($position)) && (!is_string($length))) {
        trigger_error(VA_Error(1));
    }

    if (strlen($fileString) < $position) {
       $fileString = $fileString . str_replace('*', chr(0), str_pad('', $position, '*'));
    }

    $position--;

    $int = MakeChar($int);
    $fileString = substr_replace($fileString, $int, $position, 1);

    return $fileString;                     // Return the new string.
}


/**
* Va-Functions.inc.php :: PutInteger()
*
* This function will put the passed integer into the file string at the given
* position. This is used for writing two byte integers. The default maximum
* value can be 32767, but if you set $unsigned to 1 then the maximum is raised to
* 65535 as long as it is a positive value.
*
* @param [type] $fileString
* @param [type] $int
* @param [type] $position
* @param integer $unsigned
* @return
*/
function PutInteger($fileString, $int, $position, $unsigned = 0) {
    if (!is_string($fileString)) {
        trigger_error(VA_Error(1));
    }
    if ((!is_int($int)) && (!is_string($int))) {
        trigger_error(VA_Error(1));
    }
    if ((!is_int($position)) && (!is_string($position))) {
        trigger_error(VA_Error(1));
    }

    if (strlen($fileString) < $position) {
       $fileString = $fileString . str_replace('*', chr(0), str_pad('', $position, '*'));
    }

    $position--;

    $int = MakeInteger($int, $unsigned);
    $fileString = substr_replace($fileString, $int, $position, 2);

    return $fileString;                     // Return the new string.
}


/**
* Va-Functions.inc.php :: PutLongInt()
*
* This function will put the passed integer into the file string at the given
* position. This is used for writing four byte integers. The maximum value
* can be 2147483647, but if you set $unsigned to 1, then the maximum is raise to
* 4294967295 as long as it is a positive value.
*
* @param [type] $fileString
* @param [type] $int
* @param [type] $position
* @param integer $unsigned
* @return
*/
function PutLongInt($fileString, $int, $position, $unsigned = 0) {
    if (!is_string($fileString)) {
        trigger_error(VA_Error(1));
    }
    if ((!is_int($int)) && (!is_string($int))) {
        trigger_error(VA_Error(1));
    }
    if ((!is_int($position)) && (!is_string($position))) {
        trigger_error(VA_Error(1));
    }

    if (strlen($fileString) < $position) {
       $fileString = $fileString . str_replace('*', chr(0), str_pad('', $position, '*'));
    }

    $position--;

    $int = MakeLongInt($int, $unsigned);
    $fileString = substr_replace($fileString, $int, $position, 4);

    return $fileString;                     // Return the new string.
}


/**
* Va-Functions.inc.php :: PutSingle()
*
* This function will put the passed single precision value into the file string
* at the given position.
*
* @param [type] $fileString
* @param [type] $int
* @param [type] $position
* @return
*/
function PutSingle($fileString, $int, $position) {
    if (!is_string($fileString)) {
        trigger_error(VA_Error(1));
    }
    if ((!is_int($int)) && (!is_string($int))) {
        trigger_error(VA_Error(1));
    }
    if ((!is_int($position)) && (!is_string($position))) {
        trigger_error(VA_Error(1));
    }

    if (strlen($fileString) < $position) {
       $fileString = $fileString . str_replace('*', chr(0), str_pad('', $position, '*'));
    }

    $position--;

    $int = MakeSingle($int);
    $fileString = substr_replace($fileString, $int, $position, 4);

    return $fileString;                     // Return the new string.
}


/**
* Va-Functions.inc.php :: MakeChar()
*
* This function will take the value given and convert it into a one byte
* integer represented by ascii characters.
*
* @param [type] $value
* @return
*/
function MakeChar($value) {
    if ((!is_int($value)) && (!is_string($value))) {
        trigger_error(VA_Error(1));
    }

    if ($value == 0) {
        return chr(0);
    }
    if ($value > 256) {
        trigger_error(VA_Error(2));
    }
    
    $a = chr($value);                       // Set the variable to the ascii character.
    
    return $a;                              // Return the character.
}


/**
* Va-Functions.inc.php :: MakeInteger()
*
* This function will take the value given and convert it into a two byte
* integer represented by ascii characters. The values can be from -32768 to
* 32767. If you set $unsigned to 1, then you can create an integer with a value
* up to 65535.
*
* @param [type] $value
* @param integer $unsigned
* @return
*/
function MakeInteger($value, $unsigned = 0) {
    if ((!is_int($value)) && (!is_string($value))) {
        trigger_error(VA_Error(1));
    }

    if ($value == 0) {
        return chr(0) . chr(0);
    }
    if ($unsigned == 0) {
        if ($value > 32767) {               // This is the variable type maximum.
            trigger_error(VA_Error(2));
        }
        if ($value < -32768) {
            trigger_error(VA_Error(2));
        }
        if ($value < 0) {
            $value = 65535 + $value + 1;
        }
    } else {
        if ($value > 65535) {               // This is the maximum that can be stored.
            trigger_error(VA_Error(2));
        }
    }
    
    $b = intval($value / 256);              // Divide the value.
    $value = $value - ($b * 256);           // Subtract the variable from the value.
    $b = chr($b);                           // Set the variable to the ascii character.

    $a = chr($value);                       // Set the variable to the ascii character.

    return $a . $b;                         // Return the ascii representation.
}


/**
* Va-Functions.inc.php :: MakeLongInt()
*
* This function will take the value given and convert it into a four byte
* long integer represented by ASCII characters. The values can be from
* -2147483648 to 2147483647. If you set $unsigned to 1, then you can create a
* long integer with a value up to 4294967295.
*
* @param [type] $value
* @param integer $unsigned
* @return
*/
function MakeLongInt($value, $unsigned = 0) {
    if ((!is_int($value)) && (!is_string($value))) {
        trigger_error(VA_Error(1));
    }

    if ($value == 0) {
        return chr(0) . chr(0) . chr(0) . chr(0);
    }
    if ($unsigned == 0) {
        if ($value > 2147483647) {          // This is the variable type maximum.
            trigger_error(VA_Error(2));
        }
        if ($value < -2147483648) {
            trigger_error(VA_Error(2));
        }
        if ($value < 0) {
            $value = 2147483647 + $value + 1;
        }
    } else {
        if ($value > 4294967295) {          // This is the maximum that can be stored.
            trigger_error(VA_Error(2));
        }
    }
    
    $d = intval($value / 16777216);         // Divide the value.
    $value = $value - ($d * 16777216);      // Subtract the variable from the value.
    $d = chr($d);                           // Set the variable to the ascii character.

    $c = intval($value / 65536);            // Divide the value.
    $value = $value - ($c * 65536);         // Subtract the variable from the value.
    $c = chr($c);                           // Set the variable to the ascii character.
    
    $b = intval($value / 256);              // Divide the value.
    $value = $value - ($b * 256);           // Subtract the variable from the value.
    $b = chr($b);                           // Set the variable to the ascii character.
    
    $a = chr($value);                       // Set the variable to the ascii character.

    return $a . $b . $c . $d;               // Return the ascii representation.
}


/**
* Va-Functions.inc.php :: MakeSingle()
*
* This function will take the floating point $value and convert it into a four
* byte ASCII representation. The maximum is 3.40282e+38.
*
* @param [type] $value
* @return
*/
function MakeSingle($value) {
    // Determine the sign.
    if ($value < 0) {
        $sign = 1;
    } else {
        $sign = 0;
    }

    // Split the number up into parts.
    $array = explode('.', $value);
    $integer = decbin($array[0]);
    $decimal = '';

    // Create the binary representation of the decimal.
    if (isset($array[1])) {
        $dec = '.' . $array[1];
        $fraction = 0;
        for ($x = 0; (($x < 10) && ($fraction != 1)) ; $x++) {
            $fraction = $dec * 2;
            if ($fraction >= 1) {
                $decimal = $decimal . '1';
                $dec = $fraction - 1;
            } else {
                $decimal = $decimal . '0';
                $dec = $fraction;
            }
        }
    }

    if ($integer > 0) {
        $exp = strlen($integer) - 1;
        $integer = substr($integer, 1);
        $float = $integer . $decimal;
    } else {
        if ($decimal != 0) {
            $exp = strpos($decimal, '1') + 1;
            $float = substr($decimal, $exp);
            $exp = '-' . $exp;
        } else {
            $exp = -127;
            $float = 0;
        }
    }

    $exp = decbin($exp + 127);
    $exp = str_pad($exp, 8, '0', STR_PAD_LEFT);

    $float = str_pad($float, 23, '0', STR_PAD_RIGHT);
    $float = $sign . $exp . $float;

    // Create the four bytes.
    $d = bindec(substr($float, 0, 8));
    $d = chr($d);
    $c = bindec(substr($float, 8, 8));
    $c = chr($c);
    $b = bindec(substr($float, 16, 8));
    $b = chr($b);
    $a = bindec(substr($float, 24, 8));
    $a = chr($a);

    $ascii = $a . $b . $c . $d;

    return $ascii;                          // Return the ascii representation.
}


/**
* Va-Functions.inc.php :: ReadConfig()
*
* This function will read a configuration file and place its variables into
* an array.
*
* The type of configuration file has comments designated by a semicolon, and
* variables are assigned like 'variable='value''.
*
* @param [type] $configFile
* @return
*/
function ReadConfig($configFile) {
    if (!is_string($configFile)) {
        trigger_error(VA_Error(1));
    }

    // If the configfile is missing, return the default values.
    if (!is_file($configFile)) {
        return FALSE;
    }

    // Read the configfile into an array.
    $file = file($configFile);              // Read the configfile.
    $cnt  = 1;

    for ($x = 0; $x < count($file); $x++) {
        $line    = trim($file[$x]);         // Read a line.
        $comment = trim(substr($line, 0, 1));
        if (($comment != ';') && ($line != '')) {
            $var[$cnt] = $line;             // Add the line into the array.
            $cnt++;
        }
    }

    if (!isset($var)) {                     // Check to see if any values were read.
        return $config;  
    }

    // The configuration is now in an array, so lets set the proper variables
    // with the information.

    for ($y = 1; $y != $cnt; $y++) {                            // Examine the array and set new values.
        $pos   = strpos($var[$y], '=');                         // Determine the position of the equal symbol.
        $desc  = trim(strtolower(substr($var[$y], 0, $pos)));   // Determine the variable description.
        $value = trim(substr($var[$y], $pos + 1));              // Determine the variable value.
        $value = str_replace(' ', '', $value);                  // Remove any spaces.
        $config[$desc] = strtolower($value);
    }

    return $config;                         // Return the configuration array.
}


/**
* Va-Functions.inc.php :: ReadAt2kConfig()
*
* This function will read an AT2k script configuration file, which usually is
* stored in the BBS's TXT directory.
*
* The configuration file has comments designated by a semicolon, and values are
* positioned based (they must appear in the file in a certain order and are not
* assigned to variables).
*
* @param [type] $configFile
* @return
*/
function ReadAt2kConfig($configFile) {
    if (!is_string($configFile)) {
        trigger_error(VA_Error(1));
    }

    // If the configfile is missing, send an error.
    if (!is_file($configFile)) {
        return FALSE;
    }

    // Read the configfile into an array.
    $file = file($configFile);              // Read the configfile.
    $cnt  = 1;

    for ($x = 0; $x < count($file); $x++) {
        $line    = trim($file[$x]);         // Read a line.
        $comment = trim(substr($line, 0, 1));
        if (($comment != ';') && ($line != '')) {
            $var[$cnt] = $line;             // Add the line into the array.
            $cnt++;
        }
    }

    return $var;
}


/**
* Va-Functions.inc.php :: FixHeartColors()
*
* This routine will take $string and replace any heart codes within the string
* with html color codes.
*
* @param [type] $string
* @return
*/
function FixHeartColors($string) {
    if (!is_string($string)) {
        trigger_error(VA_Error(1));
    }

    // Check to see if the variable was passed, if not then return the string.
    if ($string == '') {
        return $string;
    }
    
    $string = str_replace(chr(3) . '0', "<font color=\"#ADAAAD\">", $string);
    $string = str_replace(chr(3) . '1', "<font color=\"#52FFFF\">", $string);
    $string = str_replace(chr(3) . '2', "<font color=\"#FFFF52\">", $string);
    $string = str_replace(chr(3) . '3', "<font color=\"#AD00AD\">", $string);
    $string = str_replace(chr(3) . '4', "<font color=\"#FFFFFF\">", $string);
    $string = str_replace(chr(3) . '5', "<font color=\"#52FF52\">", $string);
    $string = str_replace(chr(3) . '6', "<font color=\"#FF5552\">", $string);
    $string = str_replace(chr(3) . '7', "<font color=\"#5255FF\">", $string);
    $string = str_replace(chr(3) . '8', "<font color=\"#AD5500\">", $string);
    $string = str_replace(chr(3) . '9', "<font color=\"#00AAAD\">", $string);
    $string = str_replace(chr(3) . 'a', "<font color=\"#00AA00\">", $string);
    $string = str_replace(chr(3) . 'b', "<font color=\"#FF55FF\">", $string);
    $string = str_replace(chr(3) . 'c', "<font color=\"#FFFFFF\">", $string);
    $string = str_replace(chr(3) . 'd', "<font color=\"#FFFF52\">", $string);
    $string = str_replace(chr(3) . 'e', "<font color=\"#52FFFF\">", $string);
    $string = str_replace(chr(3) . 'f', "<font color=\"#ADAAAD\">", $string);
    $string = str_replace(chr(3) . 'g', "<font color=\"#5255FF\">", $string);
    $string = str_replace(chr(3) . 'h', "<font color=\"#FF5552\">", $string);
    $string = str_replace(chr(3) . 'i', "<font color=\"#5255FF\">", $string);
    $string = str_replace(chr(3) . 'j', "<font color=\"#FFFF52\">", $string);
    $string = str_replace(chr(3) . 'k', "<font color=\"#52FFFF\">", $string);
    $string = str_replace(chr(3) . 'l', "<font color=\"#ADAAAD\">", $string);
    $string = str_replace(chr(3) . 'm', "<font color=\"#FFFFFF\">", $string);
    $string = str_replace(chr(3) . 'n', "<font color=\"#ADAAAD\">", $string);
    $string = str_replace(chr(3) . 'o', "<font color=\"#ADAAAD\">", $string);
    $string = str_replace(chr(3) . 'p', "<font color=\"#AD0000\">", $string);
    $string = str_replace(chr(3) . 'q', "<font color=\"#0000AD\">", $string);
    $string = str_replace(chr(3) . 'r', "<font color=\"#FFFFFF\">", $string);
    $string = str_replace(chr(3) . 's', "<font color=\"#FF5552\">", $string);
    $string = str_replace(chr(3) . 't', "<font color=\"#5255FF\">", $string);
    $string = str_replace(chr(3) . 'u', "<font color=\"#FFFFFF\">", $string);
    $string = str_replace(chr(3) . 'v', "<font color=\"#ADAAAD\">", $string);
    $string = str_replace(chr(3) . 'w', "<font color=\"#ADAAAD\">", $string);
    $string = str_replace(chr(3) . 'x', "<font color=\"#ADAAAD\">", $string);
    $string = str_replace(chr(3) . 'y', "<font color=\"#FFFFFF\">", $string);
    $string = str_replace(chr(3) . 'z', "<font color=\"#FFFFFF\">", $string);
    $string = str_replace(chr(3) . 'A', "<font color=\"#00AA00\">", $string);
    $string = str_replace(chr(3) . 'B', "<font color=\"#FF55FF\">", $string);
    $string = str_replace(chr(3) . 'C', "<font color=\"#FFFFFF\">", $string);
    $string = str_replace(chr(3) . 'D', "<font color=\"#FFFF52\">", $string);
    $string = str_replace(chr(3) . 'E', "<font color=\"#52FFFF\">", $string);
    $string = str_replace(chr(3) . 'F', "<font color=\"#ADAAAD\">", $string);
    $string = str_replace(chr(3) . 'G', "<font color=\"#5255FF\">", $string);
    $string = str_replace(chr(3) . 'H', "<font color=\"#FF5552\">", $string);
    $string = str_replace(chr(3) . 'I', "<font color=\"#5255FF\">", $string);
    $string = str_replace(chr(3) . 'J', "<font color=\"#FFFF52\">", $string);
    $string = str_replace(chr(3) . 'K', "<font color=\"#52FFFF\">", $string);
    $string = str_replace(chr(3) . 'L', "<font color=\"#ADAAAD\">", $string);
    $string = str_replace(chr(3) . 'M', "<font color=\"#FFFFFF\">", $string);
    $string = str_replace(chr(3) . 'N', "<font color=\"#ADAAAD\">", $string);
    $string = str_replace(chr(3) . 'O', "<font color=\"#ADAAAD\">", $string);
    $string = str_replace(chr(3) . 'P', "<font color=\"#AD0000\">", $string);
    $string = str_replace(chr(3) . 'Q', "<font color=\"#0000AD\">", $string);
    $string = str_replace(chr(3) . 'R', "<font color=\"#FFFFFF\">", $string);
    $string = str_replace(chr(3) . 'S', "<font color=\"#FF5552\">", $string);
    $string = str_replace(chr(3) . 'T', "<font color=\"#5255FF\">", $string);
    $string = str_replace(chr(3) . 'U', "<font color=\"#FFFFFF\">", $string);
    $string = str_replace(chr(3) . 'V', "<font color=\"#ADAAAD\">", $string);
    $string = str_replace(chr(3) . 'W', "<font color=\"#ADAAAD\">", $string);
    $string = str_replace(chr(3) . 'X', "<font color=\"#ADAAAD\">", $string);
    $string = str_replace(chr(3) . 'Y', "<font color=\"#FFFFFF\">", $string);
    $string = str_replace(chr(3) . 'Z', "<font color=\"#FFFFFF\">", $string);
    $string = str_replace(chr(3) . '[', "<font color=\"#525552\">", $string);
    $string = $string . '</font>';

    return $string;                         // Return the new string.
}


/**
* Va-Functions.inc.php :: StripBBSCodes()
*
* This routine will take $string and remove any heart codes within the string.
* It also removes diamond codes.
*
* @param [type] $string
* @return
*/
function StripBBSCodes($string) {
    if (!is_string($string)) {
        trigger_error(VA_Error(1));
    }

    // Check to see if the variable was passed, if not then return the string.
    if ($string == '') {
        return $string;
    }

    // Search for SBBS codes.
    $string = preg_replace("/\x01\S/", '', $string);

    // Search for hearts.
    $string = preg_replace("/\x03\S/", '', $string);

    // Search the file for diamonds.
    $string = preg_replace("/\x04\S/", '', $string);

    // Search for Cnet codes.
    $string = preg_replace("/\x25\S\S/", '', $string);

    // Search for pipe codes.
    $string = preg_replace("/\|(\d){2,2}/", '', $string);

    // Search for PcBoard codes.
    $string = preg_replace("/@X\S\S/", '', $string);

    // Search for Wildcat codes.
    $string = preg_replace("/@\S\S@/", '', $string);

    // Search for Celerity codes.
    $string = preg_replace("/\|[k|b|g|c|r|m|y|w|d|B|G|C|R|M|Y|W|S]/", '', $string);

    return $string;                         // Return the new string.
}


/**
* Va-Functions.inc.php :: StripANSICodes()
*
* This function will strip all standard ANSI codes from the passed string.
*
* @param [type] $string
* @return
*/
function StripANSICodes($string) {
    if (!is_string($string)) {
        trigger_error(VA_Error(1));
    }

    // Check to see if the variable was passed, if not then return the string.
    if ($string == '') {
        return $string;
    }

    $pattern = array(
                    "/\e\[(\d){0,3}[a|A|b|B|c|C|d|D]/",
                    "/\e\[((\?*?(\d){1,2})|((\d){1,2};{0,1})){0,2}[f|F|r|R]/",
                    "/\e\[((\?*?(\d){1,2})|((\d){1,2};{0,1})){0,2}[h|H]/",
                    "/\e\[(\d){0,1}[j|J]/",
                    "/\e\[(((\d){1,2};{0,1}){0,4})[m|M]/",
                    "/\e\[6[n|N]/",
                    "/\e\[((\d){1,2};{0,1}){0,3}[p|P]/",
                    "/\e\[[k|K|s|S|u|U]/"
                    );

    return preg_replace($pattern, '', $string); // Return the new string.
}


/**
* Va-Functions.inc.php :: StripRouting()
*
* This function will take $string and remove the lines that begin with a
* diamond code.
*
* @param [type] $string
* @return
*/
function StripRouting($string) {
    if (!is_string($string)) {
        trigger_error(VA_Error(1));
    }

    // Check to see if the variable was passed, if not then return the string.
    if ($string == '') {
        return $string;
    }

    // Search for diamonds.
    for ($position = 0; is_int($position);) {
        $position = strpos($string, chr(4), 0);
        if (is_int($position)) {
            $endPosition = strpos($string, "\n", $position);
            if ($endPosition === FALSE) {
                $endPosition = strlen($string);
            }
            $string = substr_replace($string, '', $position, ($endPosition - $position));
        }
    }

    return rtrim($string);                  // Return the new string.
}


/**
* Va-Functions :: ConvertLongFilename()
*
* This function will convert a long filename into a DOS 8.3 filename.
* You can pass the directory the filename will reside in ($dir), and if it
* detects multiple files with the same name it can rename it without disrupting
* the exiting files.
*
* @param [type] $filename
* @param string $dir
* @return
*/
function ConvertLongFilename($filename, $dir = '') {
    if (!is_string($filename)) {
        trigger_error(VA_Error(1));
    }

    $filename = str_replace(' ', '_', $filename);   // Replace any spaces with an underscore.

    $pos = strrpos($filename, '.');         // Find the period.
    if (($pos == 0) || ($pos === FALSE)) {  // If there is no extension, then set
        $pos = strlen($filename);           // it to the end of the file.
    }
    
    $file = substr($filename, 0, $pos);     // Get the filename.
    if ($pos < strlen($filename)) {
        $ext = substr($filename, $pos + 1); // Get the extension.
        if (strlen($ext) > 3) {
            $ext = substr($ext, 0, 3);
        }
    } else {
        $ext = '';
    }
    
    // If the filename is 8.3 compliant create the filename variable.
    if (strlen($file) <= 8) {
        $filename = $file . '.' . $ext;
    } else {
        $file = substr($file, 0, 6);
        $filename = $file . '~1.' . $ext;
    }

    if ($dir != '') {
        $path = str_replace('/', "\\", $dir);
        $filepath = $path . '/' . $filename;
        if (is_file($filepath)) {           // If that file already exists...
            $file = substr($file, 0, 6);
            $filename = $file . '.' . $ext;
            $filepath = $path . '/' . $filename;
            for ($x = 1; (is_file($filepath) && ($x < 99)); $x++) {
                if ($x == 10) {
                    $file = substr($file, 0, 5);
                }
                $filename = $file . '~' . $x . '.' . $ext;
                $filepath = $path . '/' . $filename;
            }
        }
    }
    
    return $filename;
}


/**
* Va-Functions.inc.php :: HtmlEncode()
*
* This function will take the passed line and add html href information for any
* links found. It will search for http://, ftp://, mailto: and telnet://.
*
* @param [type] $string
* @return
*/
function HTMLEncode($string) {
    if (!is_string($string)) {
        trigger_error(VA_Error(1));
    }

    $string = ' ' . $string;
    $string = preg_replace(array(
                                 "/( )(([\w\.\-]+))(@)([\w\.\-]+)\b/i",
                                 "/(([f|F][t|T][p|P](7?):\/\/)|([f|F][t|T][p|P]\.))([\w\.\/\&\=\?\-~:]+)/",
                                 "/([h|H][t|T][t|T][p|P](s?):\/\/)([\w\.\/\&\=\?\-~:#]+)/",
                                 "/( )([w|W][w|W][w|W]\.)([\w\.\/\&\=\?\-~:#]+)/",
                                 "/([t|T][e|E][l|L][n|N][e|E][t|T]:\/\/)([\w\.\/\&\=\?\-:]+)/"
                                ),
                           array(
                                 " <a href=\"mailto:$3$4$5\"\">$3$4$5</a>",
                                 "<a href=\"ftp$3://$4$5\" target=\"_blank\">$2$4$5</a>",
                                 "<a href=\"http$2://$3$4\" target=\"_blank\">$1$3$4</a>",
                                 " <a href=\"http://$2$3\" target=\"_blank\">$2$3</a>",
                                 "<a href=\"telnet://$2$3\">$1$2$3</a>"
                                ),
                                $string);

    return substr($string, 1);              // Return the encoded string, remove that inital space.
}


/**
* Va-Functions.inc.php :: ProgramTime()
*
* This function simply returns the current timestamp in microseconds.
*
* @return
*/
function ProgramTime() {
    $time = microtime();
    $time = explode(' ', $time);
    return $time[0] + $time[1];
}


/**
* Va-Functions.inc.php :: ConvertDateTime()
*
* This function will convert the many formats of dates and times into a
* standard format that VADV-PHP will use.
*
* @param [type] $string
* @return
*/
function ConvertDateTime($string) {
    $date['date'] = 'Date';
    $date['time'] = 'Time';
    if ($string == '') {
        return $date;
    }
    
    $string    = trim($string);
    $timeStamp = strtotime($string);

    if ($timeStamp == -1) {
        $pos  = strpos($string, '19');
        $year = intval(substr($string, $pos, 4));
        while (($year < 1900) && ($pos !== FALSE)) {
            $pos = strpos($string, '19', $pos + 1);
            $year = intval(substr($string, $pos, 4));
        }
        if (($year < 1970) && ($year >= 1900)) {
            $newstring = substr_replace($string, $year + 100, $pos, 4);
            $timeStamp = strtotime($newstring);
        } else {
            $a = substr($string, 0, 10);
            $b = substr($string, 10, 9);
            $c = substr($string, 19, 5);
            $newstring = $a . $c . $b;
            $timeStamp = strtotime($newstring);
        }
        if ($timeStamp == -1) {
            return $date;
        }
    }

    $date['date'] = trim(date("M j, Y", $timeStamp));
    $date['time'] = trim(date("g:i A", $timeStamp));
    
    return $date;
}


/**
* Va-Functions.inc.php :: GetAccFlags()
*
* This function will take the long integer value of the access flags and
* convert it into a string format. The returned letters are all uppercase.
*
* @param [type] $accFlagsInt
* @return
*/
function GetAccFlags($accFlagsInt) {
    if ((!is_int($accFlagsInt)) && (!is_string($accFlagsInt))) {
        trigger_error(VA_Error(1));
    }

    // Check to see if the variable was passed, if not then return nothing.
    if ($accFlagsInt == '') {
        return '';
    }
    
    $flags  = '';                           // Flags start off as none.
    $num    = 33554432;                     // This is letter Z's value as a user flag.

    for ($letter = 90; $num >= 1; $letter--) {  // It uses capital letters (letter 90).
        $tmp = intval($accFlagsInt / $num);
        if ($tmp >= 1) {
            $flags       = $flags . chr($letter);
            $accFlagsInt = $accFlagsInt - $num;
        }
        $num = $num / 2;
    }

    return strrev($flags);                  // Return the access flags.
}


/**
* Va-Functions.inc.php :: MakeAccFlags()
*
* This function will take the string representation of user access flags and
* convert it into an integer suitable for writing to userfile.dat
*
* @param [type] $accFlags
* @return
*/
function MakeAccFlags($accFlags) {
    if ((!is_int($accFlags)) && (!is_string($accFlags))) {
        trigger_error(VA_Error(1));
    }

    $accFlags    = strtoupper(str_replace(' ', '', $accFlags)); // Remove any white space and convert it to upper case.
    $accFlagsInt = 0;
    
    for ($x = strlen($accFlags); $x >= 0; $x--) {
        $letter = substr($accFlags, $x, 1); // Get a character.
        if (((ord($letter) >= 65) &&                            // It must be a letter, or ignore it.
             (ord($letter) <= 90)) &&
            (strpos($accFlags, $letter) === $x)) {              // If the letter does not occur before this one...
            $accFlagsInt = $accFlagsInt + pow(2, ord($letter) - 65);
        }
    }
    
    return $accFlagsInt;                    // Return the integer.
}


/**
* Va-Functions.inc.php :: MakeFlags()
*
* This function will take the string $accFlags and convert it to the 26 character
* representation which is writable to userfile.dat.
*
* @param [type] $accFlags
* @return
*/
function MakeFlags($accFlags) {
    if ((!is_int($accFlags)) && (!is_string($accFlags))) {
        trigger_error(VA_Error(1));
    }

    $accFlags = strtoupper(str_replace(' ', '', $accFlags));    // Remove any white space and convert it to upper case.
    $newFlags = str_pad('', 26, ' ');                           // Create the blank string.

    // Check to see if the variable was passed, if not then return nothing.
    if ($accFlags == '') {
        return $newFlags;
    }
    
    for ($x = strlen($accFlags); $x >= 0; $x--) {
        $letter = substr($accFlags, $x, 1); // Get a character.
        if ((ord($letter) >= 65) && (ord($letter) <= 90)) {     // It must be a letter, or ignore it.
            $newFlags = substr_replace($newFlags, $letter, ord($letter) - 65, 1);
        }
    }
    
    return $newFlags;                       // Return the new string.
}


/**
* Va-Functions.inc.php :: UpdateWhosOnline()
*
* This function will maintain the who's online listing for VADV-PHP. It is
* automatically called by va-functions.inc.php.
*
* @return
*/
function UpdateWhosOnline() {
    global $VA_SYSPATHS;                    // Include global variable.
    global $USER;

    $file = $VA_SYSPATHS['data'] . '/WEBCALLER.LST';    // Database filename.
    $fileString = ReadCfgFile($file);       // Read the file.

    // Read through the file looking for the current user's IP, and also
    // expired users. Users expire after 15 mins (900s).
    $user   = array();
    $online = array();
    $offset = 0;
    for ($x = 1; $offset < strlen($fileString); $x++) {
        $user['session'] = trim(GetString($fileString, 1 + $offset, 32));
        $user['ip']      = trim(GetString($fileString, 33 + $offset, 15));
        $user['time']    = intval(trim(GetString($fileString, 48 + $offset, 12)));
        $user['userno']  = intval(trim(GetInteger($fileString, 60 + $offset)));
        $user['handle']  = trim(GetString($fileString, 62 + $offset, 40));

        // Compare values to see who to remove.
        if (($user['session'] == session_id()) ||
            (time() - $user['time'] > 900) ||
            (($USER['userno'] != 0) &&
             (($user['userno'] == $USER['userno']) ||
              ($user['handle'] == $USER['handle'])))) {
            $fileString = substr_replace($fileString, '', $offset, 101);
            $x--;
        } else {
            // Write to the new fileString...
            $offset = 101 * $x;
            $online[$x]['session'] = trim($user['session']);
            $online[$x]['ip']      = trim($user['ip']);
            $online[$x]['time']    = trim($user['time']);
            $online[$x]['userno']  = trim($user['userno']);
            $online[$x]['handle']  = trim($user['handle']);
        }
    }

    // Add the current user's entry to the end if not a web spider.
    if (!DetectSpider()) {
        $sess   = str_pad(session_id(), 32);
        $ip     = str_pad($_SERVER['REMOTE_ADDR'], 15);
        $time   = str_pad(time(), 12);
        $userno = MakeInteger($USER['userno']);
        $handle = str_pad($USER['handle'], 40);

        $fileString = $fileString . $sess . $ip . $time . $userno . $handle;

        $online[$x]['session'] = trim($sess);
        $online[$x]['ip']      = trim($ip);
        $online[$x]['time']    = trim($time);
        $online[$x]['userno']  = trim($USER['userno']);
        $online[$x]['handle']  = trim($USER['handle']);
    } else {
        $x--;
    }

    // Write the new file.
    WriteCfgFile($file, $fileString);

    // Organize the array alphabetically by handle.
    if (count($online) > 0) {
        foreach ($online as $row) {
            $sortarr[] = $row['handle'];
        }
        array_multisort($sortarr, $online);
        array_unshift($online, $x);
    } else {
        $online[0] = 0;
    }

    return $online;                         // Return the online users.
}


/**
* Va-Functions.inc.php :: WriteLog()
*
* This function will log a line into the specified file which is stored in
* the BBS's data directory. The logfile is a filename. MaxSize is an optional
* parameter which can be used to reset the log file if it is larger than the
* value (in bytes).
*
* @param [type] $logFilename
* @param [type] $line
* @param integer $maxSize
* @return
*/
function WriteLog($logFilename, $line, $maxSize = 0) {
    global $VA_SYSPATHS;

    $logFile = $VA_SYSPATHS['data'] . '/' . $logFilename;

    // If log is larger than $maxSize then reset it.
    if (($maxSize != 0) && (is_file($logFile)) && (filesize($logFile) >= $maxSize)) {
        $oldFile = current(explode('.', basename($logFilename)));
        $oldFile = $VA_SYSPATHS['data'] . '/' . $oldFile . '.OLD';  // End the old file in .old
        if (is_file($oldFile)) {                                    // If a previous old file exists, delete it.
            unlink($oldFile);
        }
        copy($logFile, $oldFile);                                   // Copy the current log.
        unlink($logFile);                                           // Delete the current log.
    }
    
    if (!is_file($logFile)) {
        $fp = fopen($logFile, 'w');
        flock($fp, LOCK_EX);
        fwrite($fp, "\r\n");
        fwrite($fp, "VADV-PHP Event Log\r\n");
        fwrite($fp, "-------------------------------------------------------------------------------\r\n");
        fwrite($fp, "Date           Time     IP Address         Description\r\n");
        fwrite($fp, "-------------------------------------------------------------------------------\r\n");
        flock($fp, LOCK_UN);
        fclose($fp);
    }

    $fp = fopen($logFile, 'a');
    flock($fp, LOCK_EX);                    // Lock the file.

    $ip = '[' . trim($_SERVER['REMOTE_ADDR']) . ']';
    $ip = str_pad($ip, 17, ' ');

    $date   = date('D') . ' ' . date('Y') . '-' . date('m') . '-' . date('d');
    $time   = date('H') . ':' . date('i') . ':' . date('s');
    $line   = $date . ' ' . $time . ' ' . $ip . '  ' . $line . "\r\n";
    $status = fwrite($fp, $line);           // Write the line to the file.

    flock($fp, LOCK_UN);                    // Unlock the file.
    fclose($fp);                            // Close the file.
    
    return $status;                         // Return the result.
 }


/**
* Va-Functions.inc.php :: SendOneLiner()
*
* This function will send a one-liner to the specified node, or to all active
* nodes. If the userNo is 0, then it will send the message to any user. If
* the node is set to all, it will send it to all active nodes.
*
* @param [type] $message
* @param integer $userNo
* @param string $node
* @return
*/
function SendOneLiner($message, $userNo = 0, $node = 'All') {
    if (!is_string($message)) {
        trigger_error(VA_Error(1));
    }
    if ((!is_int($userNo)) && (!is_string($userNo))) {
        trigger_error(VA_Error(1));
    }

    global $VA_SYSPATHS;                                // Include global variable.
    global $includeDir;
    include_once($includeDir . 'va-control.inc.php');   // Include control.dat functions.
    include_once($includeDir . 'va-main.inc.php');      // Include main configuration.
    global $VA_MAIN;
    
    // Truncate the message if too long.
    if (strlen($message) > 75) {
        $message = substr($message, 0, 75);
    }
    // Add coloring.
    $message = chr(3) . '6*** ' . chr(3) . '1' . $message;

    $status = FALSE;
    if (($userNo == 0) && ($node != 'All')) {
        $file = $VA_SYSPATHS['control'] . '/in' . $node . '.que';
        $fp = fopen($file, 'a');            // Open the file.
        flock($fp, LOCK_EX);                // Lock the file.
        $status = fwrite($fp, $message);    // Write the message.
        flock($fp, LOCK_UN);                // Unlock the file.
        fclose($fp);
    } else {
        if ($userNo > 0) {
            for ($x = 0; $x < $VA_MAIN['maxchannelno']; $x++) {
                $channel = GetControl($x);
                if ($channel['userno'] == $userNo) {
                    $file = $VA_SYSPATHS['control'] . '/in' . $x . '.que';
                    $fp = fopen($file, 'a');            // Open the file.
                    flock($fp, LOCK_EX);                // Lock the file.
                    $status = fwrite($fp, $message);    // Write the message.
                    flock($fp, LOCK_UN);                // Unlock the file.
                    fclose($fp);
                }
            }
        } else {
            for ($x = 0; $x < $VA_MAIN['maxchannelno']; $x++) {
                $channel = GetControl($x);
                if ($channel['userno'] > 0) {
                    $file = $VA_SYSPATHS['control'] . '/in' . $x . '.que';
                    $fp = fopen($file, 'a');            // Open the file.
                    flock($fp, LOCK_EX);                // Lock the file.
                    $status = fwrite($fp, $message);    // Write the message.
                    flock($fp, LOCK_UN);                // Unlock the file.
                    fclose($fp);
                }
            }
        }
    }

    return $status;                         // Return the status.
}


/**
* Va-Functions.inc.php :: GetCurrentURL()
*
* This function simply returns the full URL of the current script. If $query is
* set to 0, then it will return the current URL without the query string.
*
* @param integer $query
* @return
*/
function GetCurrentURL($query = 1) {
    $url = 'http://' . $_SERVER['HTTP_HOST'] . $_SERVER['SCRIPT_NAME'];
    if (strpos($url, 'php.exe') !== FALSE) {
        $url = 'http://' . $_SERVER['HTTP_HOST'] . $_SERVER['PHP_SELF'];
    }
    if ((isset($_SERVER['QUERY_STRING'])) &&
        ($query != 0) &&
        (trim($_SERVER['QUERY_STRING']) != '')) {
        $url = $url . '?' . $_SERVER['QUERY_STRING'];
    }

    return $url;                            // Return the url.
}


/**
* Va-Funtions.inc.php :: RedirectUser()
*
* This function will redirect the user to the specified URL. It uses a primary
* base of JavaScript with a backup meta tag for those browsers that do not
* support JavaScript. Note that when this is called, the current page will
* exit and the redirection will take place. You may set delay the redirection
* a specific number of seconds by using the optional $delay variable.
*
* @param [type] $url
* @param integer $delay
* @return
*/
function RedirectUser($url, $delay = 0) {
    if (!is_string($url)) {
        return FALSE;
    }
    $delay = intval($delay);

    if ((!headers_sent()) &&
        (strpos($url, 'http://') !== FALSE) &&
        (strpos(strtoupper($_SERVER['SERVER_SOFTWARE']), 'IIS') === FALSE) &&
        ($delay < 1)) {
        // header('Refresh: ' . $delay . '; URL=' . $url);
        header('Location: ' . $url);
    } else {
?>
<html>

<head>
    <title></title>
    <noscript>
        <meta http-equiv="Refresh" content="<?php print $delay; ?>; url=<?php print $url; ?>" />
    </noscript>

    <script language="JavaScript">
    <!--
    function doRedirect() {
        setTimeout("window.location.href='<?php print $url; ?>'", <?php print $delay; ?>*1000);
    }
    //-->
    </script>
</head>

<body onload="doRedirect()"></body>

</html>
<?php
    }

    die();
}


/**
* Va-Functions.inc.php :: DetectSpider()
*
* This function detects if the current user client is a known search engine
* spider. If so, it returns TRUE, else it returns FALSE.
*
* @return
*/
function DetectSpider() {
    // The following search engine identifier information was obtained from:
    // http://www.jafsoft.com/searchengines/webbots.html
    if ((isset($_SERVER['HTTP_USER_AGENT'])) &&
        ((stristr($_SERVER['HTTP_USER_AGENT'], 'AbachoBOT')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'abcdatos_botlink')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'AESOP_com_SpiderMan')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'ah-ha.com')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'ia_archiver')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Scooter')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Mercator')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Scooter2_Mercator_3-')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'roach.smo.av.com-')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Tv<nn>_Merc_resh_26_1_D-')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'AltaVista')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'FAST-WebCrawler')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Wget')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Acoon Robot')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'antibot')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Atomz')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Buscaplus Robi')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'CanSeek')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'ChristCRAWLER')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Crawler')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'RaBot')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'DeepIndex')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'DittoSpyder')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Jack')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Speedy Spider')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'ArchitextSpider')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'ArchitectSpider')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Arachnoidea')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'EZResult')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Fast PartnerSite')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'FAST Data Search')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'KIT-Fireball')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'GalaxyBot')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'geckobot')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'GenCrawler')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Googlebot')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'moget')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Aranha')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Slurp')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Toutatis')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Hubater')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'IlTrovatore-Setaccio')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'IncyWincy')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'UltraSeek')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'InfoSeek')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Mole2')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'MP3Bot')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'ip3000')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'kuloko-bot')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'LNSpiderguy')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'NetResearchServer')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'MantraAgent')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Lycos')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'HenryTheMiragoRobot')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'MSNBOT')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Gulliver')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'PicoSearch')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'PJspider')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'DIIbot')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'nttdirectory_robot')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'griffon')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'maxbot')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'gazz')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'NationalDirectory-SuperSpider')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'naver')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Openfind')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Openbot')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'psbot')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'CrawlerBoy')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'AlkalineBOT')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Fluffy')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Scrubby')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'asterias')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'speedfind')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Kototoi')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'SightQuestBot')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Surfnomore')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'teoma')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'ESISmartSpider')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Tutorial Crawler')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'UK Searcher')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Vivante')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Nazilla')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'WebWombat')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'marvin')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'MuscatFerret')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'WhizBang')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'ZyBorg')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'WIRE WebRefiner')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'WSCbot')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Yandex')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Yellopet-Spider')) ||
         (stristr($_SERVER['HTTP_USER_AGENT'], 'Iron33')))) {
        // It is a spider...
        return TRUE;
    } else {
        // Is not a spider...
        return FALSE;
    }
}


/**
* Va-Functions.inc.php :: Array_StripSlashes()
*
* This function will remove slashes from an array.
*
* @param [type] &$array
* @return
*/
function Array_StripSlashes($array) {
    if (!is_array($array)) {
        return $array;
    }
    while (list($key) = each($array)) {
        if (is_array($array[$key])) {
            array_stripslashes($array[$key]);
        } else {
            $array[$key] = stripslashes($array[$key]);
        }
    }
    
    return $array;
}


/**
* Va-Functions.inc.php :: SendFile()
*
* This function will send the specified file to the user using the fpassthru()
* function.
*
* @param [type] $file
* @return
*/
function SendFile($file) {
    if (!is_file($file)) {
        return FALSE;
    }

    $status     = FALSE;                    // Set the status to default to false.
    $filename   = basename($file);          // Get the filename.
    $filesize   = filesize($file);          // Get the filesize.
    $filesizemb = intval($filesize / 1048576);

    // Determine MIME type.
    $pos = strrpos($filename, '.');
    $ext = strtolower(substr($filename, $pos));
    switch ($ext) {
        case '.ai':   $mimetype='application/postscript';   $encode='7bit';     break;
        case '.avi':  $mimetype='video/x-msvideo';          $encode='base64';   break;
        case '.doc':  $mimetype='application/msword';       $encode='base64';   break;
        case '.eps':  $mimetype='application/postscript';   $encode='7bit';     break;
        case '.exe':  $mimetype='application/octet-stream'; $encode='base64';   break;
        case '.gif':  $mimetype='image/gif';                $encode='base64';   break;
        case '.htm':  $mimetype='text/html';                $encode='7bit';     break;
        case '.html': $mimetype='text/html';                $encode='7bit';     break;
        case '.jpe':  $mimetype='image/jpeg';               $encode='base64';   break;
        case '.jpeg': $mimetype='image/jpeg';               $encode='base64';   break;
        case '.jpg':  $mimetype='image/jpeg';               $encode='base64';   break;
        case '.mpe':  $mimetype='video/mpeg';               $encode='base64';   break;
        case '.mpeg': $mimetype='video/mpeg';               $encode='base64';   break;
        case '.mov':  $mimetype='video/quicktime';          $encode='base64';   break;
        case '.mpg':  $mimetype='video/mpeg';               $encode='base64';   break;
        case '.pdf':  $mimetype='application/pdf';          $encode='base64';   break;
        case '.ps':   $mimetype='application/postscript';   $encode='7bit';     break;
        case '.qt':   $mimetype='video/quicktime';          $encode='base64';   break;
        case '.rtf':  $mimetype='application/rtf';          $encode='7bit';     break;
        case '.tar':  $mimetype='application/octet-stream'; $encode='base64';   break;
        case '.tif':  $mimetype='image/tiff';               $encode='base64';   break;
        case '.tiff': $mimetype='image/tiff';               $encode='base64';   break;
        case '.txt':  $mimetype='text/plain';               $encode='7bit';     break;
        case '.wav':  $mimetype='audio/x-wav';              $encode='base64';   break;
        case '.zip':  $mimetype='application/zip';          $encode='base64';   break;
        default:      $mimetype='application/octet-stream'; $encode='base64';   break;
    }

    // Set some new time limits and file sizes. (Do not think this is necessary)
    ini_set('max_execution_time', 240);
    ini_set('memory_limit', 4 + $filesizemb . 'M');

    header('Cache-Control: ');
    header('Pragma: ');
    header('Content-type: ' . $mimetype);
    header('Content-Disposition: attachment; filename="' . $filename . '"');
    header('Content-transfer-encoding: ' . $encode);
    header('Content-length: ' . $filesize);

    $fp = fopen($file, 'rb');               // Open the file to send.
    $status = fpassthru($fp);               // Send the file.

    return $status;                         // Return the status.
}


/**
* Va-Functions.inc.php :: GlobalFooter()
*
* A footer for all pages. Displays execution time and copyright information.
* For dark backgrounds, set $color to 1.
*
* @param integer $color
* @param integer $noBar
* @return
*/
function GlobalFooter($color = 0, $noBar = 0) {
    global $CONFIG;
    global $VERSION;
    global $startTime;

    $totalTime = round(ProgramTime() - $startTime, 4);
    if ($color == 0) {
?>

<style>
a.footerBoldBody:link    { font-family: Verdana, Arial, Helvetica; font-size: 8pt;  color: #000000; font-weight:bold }
a.footerBoldBody:visited { font-family: Verdana, Arial, Helvetica; font-size: 8pt;  color: #000000; font-weight:bold }
.footerBody              { font-family: Verdana, Arial, Helvetica; font-size: 8pt;  color: #000000 }
.footerBoldBody          { font-family: Verdana, Arial, Helvetica; font-size: 8pt;  color: #000000; font-weight:bold }
.footerVersion           { font-family: Courier New;               font-size: 8pt;  color: #808080; font-weight:bold }
</style>

<?php
if ($noBar == 0) {
    print '<hr color="#000080">';
} else {
    print '<br />';
}
?>
<div align="center">
    <center>
    <table bgcolor="#FFFFFF" border="0" cellpadding="5" cellspacing="0" style="border-collapse: collapse" bordercolor="#111111" width="100%">
        <tr>
            <td width="100%" valign="top">
                <p align="center"><span class="footerBoldBody">This page is powered by <a class="footerBoldBody" href="http://www.vadvphp.com">VADV-PHP</a></span><br /><span class="footerBody">Execution Time: <?php print $totalTime; ?> seconds</span></p>
                <p align="center" class="footerBody">If you experience any problems with this website or need help, contact the <a href="mailto:<?php print $CONFIG['webmaster']; ?>">webmaster</a>.<br /> VADV-PHP Copyright  2002-<?php print date('Y'); ?> Steve Winn. All Rights Reserved.<br /> Virtual Advanced Copyright  1995-1997 Roland De Graaf.<br /><span class="footerVersion">v<?php print $VERSION; ?></span></p>
            </td>
        </tr>
    </table>
    </center>
</div>

<?php
    } else {
?>

<style>
a.footerBoldBody:link    { font-family: Verdana, Arial, Helvetica; font-size: 8pt;  color: #FFFFFF; font-weight:bold }
a.footerBoldBody:visited { font-family: Verdana, Arial, Helvetica; font-size: 8pt;  color: #FFFFFF; font-weight:bold }
.footerBody              { font-family: Verdana, Arial, Helvetica; font-size: 8pt;  color: #FFFFFF }
.footerBoldBody          { font-family: Verdana, Arial, Helvetica; font-size: 8pt;  color: #FFFFFF; font-weight:bold }
.footerVersion           { font-family: Courier New;               font-size: 8pt;  color: #C0C0C0; font-weight:bold }
</style>

<hr color="#525552">
<div align="center">
    <center>
    <table bgcolor="#000000" border="0" cellpadding="5" cellspacing="0" style="border-collapse: collapse" bordercolor="#111111" width="100%">
        <tr>
            <td width="100%" valign="top">
                <p align="center"><span class="footerBoldBody">This page is powered by <a class="footerBoldBody" href="http://www.vadvphp.com">VADV-PHP</a></span><br /><span class="footerBody">Execution Time: <?php print $totalTime; ?> seconds</span></p>
                <p align="center" class="footerBody">If you experience any problems with this website or need help, contact the <a href="mailto:<?php print $CONFIG['webmaster']; ?>">webmaster</a>.<br /> VADV-PHP Copyright  2002-<?php print date('Y'); ?> Steve Winn. All Rights Reserved.<br /> Virtual Advanced Copyright  1995-1997 Roland De Graaf.<br /><span class="footerVersion">v<?php print $VERSION; ?></span></p>
            </td>
        </tr>
    </table>
    </center>
</div>

<?php
    }
}


/**
* Va-Functions.inc.php :: VA_Die()
*
* This function replaces the php die() command. It will display an error message
* for five seconds, and then if the directory contains an index.htm, index.html
* or index.php file, it will redirect to it.
*
* @param [type] $message
* @return
*/
function VA_Die($message, $delay = 5) {
    $index = '';
    if (is_file('index.php')) {
        $index = 'index.php';
    } else if (is_file('index.htm')) {
        $index = 'index.htm';
    } else if (is_file('index.html')) {
        $index = 'index.html';
    }

    if ($index != '') {
        print $message;
        RedirectUser($index, $delay);
    } else {
        die($message);
    }
}


/**
* Va-Functions.inc.php :: VA_Error()
*
* This function will return the specified error message.
*
* @param [type] $errNum
* @return
*/
function VA_Error($errNum) {
    if ((!is_int($errNum)) && (!is_string($errNum))) {
        $errNum = 4;
    }

    switch ($errNum) {
        case 0:
            $errMsg = 'File not found.';
            break;
        case 1:
            $errMsg = 'Variable is not proper type.';
            break;
        case 2:
            $errMsg = 'Value is out of range.';
            break;
        case 3:
            $errMsg = 'Invalid Configuration.';
            break;
        case 4:
            $errMsg = 'Unknown Error.';
            break;
    }
    
    return $errMsg;                         // Return the error message.
}


/**
* Va-Functions.inc.php :: VA_LoginError()
*
* This function will display a message stating that the user needs to login to
* access a certain area and give a link to the generic login page.
*
* @return
*/
function VA_LoginError() {
    global $CONFIG;                         // Include global variable.
    global $includeDir;
?>
<html>

<head>
<title>Login Authentication Error</title>
<meta http-equiv="Content-Type" content="text/html; charset=ISO-8559-1" />
<meta name="copyright" content="Copyright  2003, Steve Winn" />
<link rel="stylesheet" type="text/css" href="./../VADVPHP.css" />
</head>

<body bgcolor="#000000">

<div align="center">
    <center>
    <table border="1" width="640" cellspacing="0" cellpadding="5" bgcolor="#000080" bordercolor="#FFFFFF">
        <tr>
            <td valign="top">
                <p class="yellowTitle" align="center">Login Authentication Error</p>
                <table cellspacing="1" width="100%">
                    <tr>
                        <td valign="top">
                            <p class="navBar" align="center">The area you are attempting to access
                            requires that you login first.</p>
                            <p align="center"><a class="return" href="<?php print $includeDir; ?>./../bbslogin.php?url=<?php print urlencode(GetCurrentURL()); ?>">Click here to login.</a></p>
                        </td>
                    </tr>
                </table>
                <p align="center"><span class="navBar">This page is powered
                by <a class="navBar" href="http://www.vadvphp.com">VADV-PHP</a></span></p>
                <p align="center" class="infoBarBody">If you experience any problems
                with this website or need help, contact the
                <a class="infoBarBody" href="mailto:<?php print $CONFIG['webmaster']; ?>">webmaster</a>.<br />
                VADV-PHP Copyright  2002-<?php print date('Y'); ?> Aspect
                Technologies. All Rights Reserved.</p>
            </td>
        </tr>
    </table>
    </center>
</div>

</body>

</html>
<?php
}


/**
* Va-Functions.inc.php :: CustomErrorHandler()
*
* This is a custom error handler in case all hell breaks loose.
*
* @param [type] $errNumber
* @param [type] $errDesc
* @param [type] $errFile
* @param [type] $errLine
* @return
*/
function CustomErrorHandler($errNumber, $errDesc, $errFile, $errLine) {
    global $VERSION;
    global $DEBUG_MAIL;
    
    $smtp_server = 'mail.at2k.org';

    // Send an email to the VADV-PHP bug report address.
    if ($DEBUG_MAIL == TRUE) {
        if (gethostbyname($smtp_server) != $smtp_server) {
            $phpVersion = phpversion();
            $currentURL = GetCurrentURL();

            $pos  = strrpos($errFile, chr(92)) + 1;
            $file = substr($errFile, $pos);

            // Create error message.
            $error = <<<EOD
Error number: $errNumber
Description:  $errDesc
Script name:  $file
Line number:  $errLine

URL:          $currentURL
PHP version:  $phpVersion
EOD;

            ini_set('SMTP', $smtp_server);
            ini_set('sendmail_from', 'bugs@vadvphp.com');
            $headers  = "MIME-Version: 1.0\n";
            $headers .= "Reply-To: bugs@vadvphp.com\n";
            $headers .= "X-Sender: Bug Report<bugs@vadvphp.com>\n";
            $headers .= "X-Mailer: PHP/" . $phpVersion . "\n";
            $headers .= "X-Priority: 3\n";
            $headers .= "Return-Path: <bugs@vadvphp.com>\n";

            mail('bugs@vadvphp.com',
                 '[NON-FATAL] VADV-PHP ERROR (' . $VERSION . ')',
                 'http://' . $_SERVER['SERVER_NAME'] . "\n" . $error,
                 $headers);
            ini_restore('SMTP');
            ini_restore('sendmail_from');
        }
    }

    // Display error message to user.
    print "<h2>Unable to display page.</h2>\n";
    print "<p>An error occurred while processing your request. The site administrator has been notified.</p>\n";
    print "<p>We apologize for the inconvenience.</p>";
    VA_Die('This page will attempt to be redirected in 5 seconds.');
}
?>
